/* get_alt.c - A function to find an altitude from a data set for a given
 *             latitude and longitude for EARTH
 *
 * Copyright (C) 2005 - 2006
 *                    Joern Wilms
 *                    Robert Dowse
 *                    Robert Swain
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any phier version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 *
 */

#include <assert.h>
#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "../druantia.h"
#include "common.h"


int get_alt(coords_t *coords, char *datapath, short *alt) {
  //
  // get altitude at geographic position phi/lam from
  // ETOPO2 data
  //
  // phi: north is positive, constrained to -90...+90
  // lam: east is positive, constrained to -180...+180
  //

  int col, row;
  long index;

  // if we haven't done so, mmap the altitude data
  static unsigned char *etopo=NULL;

  if(isnan(coords->phi) || isnan(coords->lam)){
    *alt = -32767;
    return(OK);
  }

  if ( !etopo ) {
    // stupid me, this is the big endian format
    int fd=open(datapath,O_RDONLY);
    if ( fd == -1 ) {
      print_errno_info(errno);
      exit(1);
    }

    etopo=mmap(NULL,116640000,PROT_READ,MAP_SHARED,fd,0);
    assert(etopo);
  }

  if (coords->phi < -90.0 || coords->phi > 90.0) {
#ifdef DEBUG
    printf("PROBLEM phi: %f\n",coords->phi);
#endif
    if (coords->phi <- 90.0) { coords->phi = -90.0;}
    if (coords->phi > 90.0) { coords->phi = 90.0;}
  }

  if (coords->lam<-180.0 || coords->lam > 180.0) {
#ifdef DEBUG
    printf("PROBLEM lam: %f\n",coords->lam);
#endif
    if (coords->lam<-180.0) { coords->lam=-180.0;}
    if (coords->lam> 180.0) { coords->lam= 180.0;}
  }

  assert(coords->phi>=-90.0 && coords->phi<=90.0);
  assert(coords->lam>=-180.0 && coords->lam<=180.0);

  // etopo covers 180deg West to 179deg 58' East and
  // 90deg North to 89deg 58' South.
  // assume 2810m for the elevation of the south pole
  // size of etopo2: 10800columns and 5400 rows

  // column
  col = (int)((coords->lam+180.)*10800./360.+0.5);
  if (col<0) { col=0; }       // take care of round off
  if (col>=10800) {col=10799;}

  // row
  row = (int)((90.-coords->phi)*5400./180.+0.5);
  if (row<0) { row=0;}

  // south pole
  if (row>=5400) {
    return 2810.;
  }

  assert(col>=0 && col<10800);
  assert(row>=0 && row<5400);

  // get index into topography file
  index = (row*10800+col)*2;

  *alt = etopo[index+1];
  *alt |= (etopo[index]<<8);

  return(OK);
}
