/* get_alt_mars.c - A function to find an altitude from a data set for a given
 *             latitude and longitude for Mars
 *
 * Copyright (C) 2005 - 2006
 *                    Joern Wilms
 *                    Robert Dowse
 *                    Robert Swain
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any phier version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 */

#include <assert.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "../druantia.h"
#include "common.h"


int get_alt_mars(coords_t *coords, char *datapath, short *alt) {
  //
  // get altitude at geographic position phi/lam from
  // Mars MOLA data
  //
  // phi: north is positive, constrained to -90...+90
  // lam: east is positive, constrained to -180...+180
  //

  int col, row;
  long index;
  double phi,lam;
  int lam0,phi0;
  char filename[17];
  static char memfile[17]="\0";
  // pointer to the memmapped data
  static unsigned char *etopo=NULL;
  static size_t memlength=-1;

  char *marspath="/storage/data1/mola/";
//  char *marspath="/data1/druantia/MOLA/pds-geosciences.wustl.edu/geodata/mgs-m-mola-5-megdr-l3-v1/mgsl_300x/meg128/";


  if(isnan(coords->phi) || isnan(coords->lam)){
    *alt = -32767;
    return(OK);
  }

  phi=coords->phi;
  lam=coords->lam;

  if (phi < -90.0 || phi > 90.0) {
    if (phi <- 90.0) { phi = -90.0;}
    if (phi > 90.0) { phi = 90.0;}
  }

  if (lam<-180.0 || lam > 180.0) {
    if (lam<-180.0) { lam=-180.0;}
    if (lam> 180.0) { lam= 180.0;}
  }

  assert(phi>=-90.0 && phi<=90.0);
  assert(lam>=-180.0 && lam<=180.0);

  if ( lam<0 ) { lam+=360; }

  //
  // get name of current tile.
  //

  phi0=(int) phi;

  // there are no data above 88deg
  if ( phi0 >=88 || phi0<=-88 ) {
    *alt=-32767;
    return(OK);
  }

//  phi0 > 44 ? phi0 = 88 : phi0 > 0 ? phi0 = 44 : phi0 > -44 ? phi0 = 0 : phi0 = -44;
  if (phi0 > 44 ) {
    phi0=88;
  } else {
    if (phi0 > 0 ) {
      phi0=44;
    } else {
      if (phi0 > -44 ) {
        phi0=0;
      } else {
        phi0=-44;
      }
    }
  }

  lam0=(int) lam;
//  lam0 > 270 ? lam0 = 270 : lam0 > 180 ? lam0 = 180 : lam0 > 90 ? lam0 = 90 : lam0 = 0;
  if (lam0 > 270 ) {
    lam0=270;
  } else {
    if (lam0 > 180 ) {
      lam0=180;
    } else {
      if (lam0 > 90 ) {
        lam0=90;
      } else {
        lam0=0;
      }
    }
  }

  // construct MOLA file name
  char sign= (phi0>=0 ) ? 'n' : 's';
  sprintf(filename,"megt%2.2i%1c%3.3ihb.img",abs(phi0),sign,lam0);

  // each Mars tile covers 90 deg in longitude and 44 deg in latitude
  // and covers 11520 columns and 5632 rows
  // (needed here for the assertion in the mmap below)
  long maxcol=11520;
  long maxrow= 5632;

  if ( !etopo || (strcmp(filename,memfile)!=0) ){
    if (etopo!=NULL) {
      // not the first mmap --> unmap existing file
      int ret=munmap(etopo,memlength);
      if ( ret!=0 ) {
        print_errno_info(errno);
        exit(1);
      }
      etopo=NULL; //  not necessary, really
    }

    // memory map the data file if it is not yet in memory
    struct stat fileinfo;

    char *datafile=(char *)malloc(strlen(marspath)+strlen(filename)+1);
    strncpy(datafile,marspath,strlen(marspath));
    strncat(datafile,filename,strlen(filename));

    stat(datafile,&fileinfo);
    int fd=open(datafile,O_RDONLY);
    if ( fd == -1 ) {
      print_errno_info(errno);
      exit(1);
    }
    assert(fileinfo.st_size == maxrow*maxcol*2);

    memlength=fileinfo.st_size;
    etopo=mmap(NULL,memlength,PROT_READ,MAP_SHARED,fd,0);
    if ( etopo == MAP_FAILED ) {
      print_errno_info(errno);
      exit(1);
    }
    // remember currently mapped file
    strncpy(memfile,filename,strlen(filename));

    // close fd again to avoid too many open files -- the mmap
    // remains!
    close(fd);
  }


  // column
  col = (int)((lam-lam0)*maxcol/90.+0.5);
  if (col<0) { col=0; }       // take care of round off
  if (col>=maxcol) {col=maxcol-1;}

  // row
  row = (int)((phi0-phi)*maxrow/44.+0.5);
  if (row<0) { row=0;}
  if (row>=maxrow) { row=maxrow-1;}

  assert(col>=0 && col<maxcol);
  assert(row>=0 && row<maxrow);

  // get index into topography file
  index = (row*maxcol+col)*2;

  assert(index>=0 && index < memlength-2);

  // get the altitude (works on all machines)
  *alt = etopo[index+1];
  *alt |= (etopo[index]<<8);

  return(OK);
}
