/* get_lam_phi_satellite.c - Convert (x,y) into (lam,phi) using the satellite
 *                           projection
 *
 * Copyright (C) 2005 - 2006
 *                    Joern Wilms
 *                    Robert Dowse
 *                    Robert Swain
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>

// get sincos function on GCC linux systems
#ifdef __GNUC__
#define _GNU_SOURCE
#endif

#include <math.h>

#include "../druantia.h"
#include "common.h"

int get_lam_phi_satellite(main_t *main, int x, int y, coords_t *coords,
      double *incart){

    double *cart;
    double x_prime, y_prime, z_prime;
    double x_prime2, y_prime2, z_prime2;
    double lam0, phi0;
    double sin_lam0, cos_lam0, sin_phi0, cos_phi0;
    double metres_per_pixel;

    if ( incart != NULL ) {
      cart = incart;
    } else {
      cart = (double *) malloc(3*sizeof(double));
      if ( cart == NULL ) return(ERROR);
    }

    // Convert the center longitude and latitude from degrees into radians.
    lam0 = (main->lam0)* M_PI / 180.0;
    phi0 = (main->phi0)* M_PI / 180.0;

    // Precompute the sine and cosine of lam0 and phi0
#ifdef _GNU_SOURCE
    sincos(lam0,&sin_lam0,&cos_lam0);
    sincos(phi0,&sin_phi0,&cos_phi0);
#else
    sin_lam0 = sin(lam0); cos_lam0 = cos(lam0);
    sin_phi0 = sin(phi0); cos_phi0 = cos(phi0);
#endif

    //  How many metres in real life that the pixel represents.
    metres_per_pixel = main->scale * INS_TO_M /main->dpi;

    //  Find the cartesian coordinates for the centre of the pixel.
    cart[1] = ( (double)x - (main->width  - 1.0)/2.0) * metres_per_pixel;
    cart[2] = ( (double)y - (main->height - 1.0)/2.0) * metres_per_pixel;
    cart[0] = sqrt( main->rad_of_planet*main->rad_of_planet - cart[1]*cart[1] - cart[2]*cart[2]  );

    //  Rotation about the y-axis towards the equator.
    x_prime = cos_phi0*cart[0] - sin_phi0*cart[2];
    y_prime = cart[1];
    z_prime = sin_phi0*cart[0] + cos_phi0*cart[2];

    //  Rotation about the z-axis from the meridian to lambda=0.
    x_prime2 =  cos_lam0*x_prime - sin_lam0*y_prime;
    y_prime2 =  sin_lam0*x_prime + cos_lam0*y_prime;
    z_prime2 =  z_prime;

    // Find the spherical coordinates (r, lam, phi) from the 
    // cartesian coordinates (x,y,z).
    coords->lam = atan2(y_prime2,x_prime2) * 180.0/M_PI;
    coords->phi = asin(z_prime2/main->rad_of_planet) * 180.0/M_PI;

    return(OK);

}
