/* druantia2d.c - A program to generate 2D images, output in PNG format, of a
 *                 planet using altitude data
 *
 * Copyright (C) 2005 - 2006
 *                    Joern Wilms
 *                    Robert Dowse
 *                    Robert Swain
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <getopt.h>
#include <math.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sysexits.h>

#include "../druantia.h"
#include "../common/common.h"
#include "../common/imglib/imglib.h"


int druantia2d(main_t *main){

  int x,y;
  short alt;
  uint8_t *bmp;
  coords_t *coords;
  pixel_t *pixel;
  bitmap_t *bitmap;

  // malloc structs first used in druantia2d()
  if((coords = (coords_t *)malloc(sizeof(coords_t))) == NULL){
    return(ERROR);
  }
  if((pixel = (pixel_t *)malloc(sizeof(pixel_t))) == NULL){
    return(ERROR);
  }
  if((bitmap = (bitmap_t *)malloc(sizeof(bitmap_t))) == NULL){
    return(ERROR);
  }

  bitmap->width = main->width;
  bitmap->height = main->height;
  bitmap->name = main->pngpath;

  //init the bitmap in memory
  if((bmp = bmp_init(bitmap)) == NULL){
    fprintf(stderr,"bitmap initialisation failed\n");
    return(ERROR);
  }


  //loop over all pixels in scanline fashion (one row at a time)
  for(y=0; y<main->height; y++){
    for(x=0; x<main->width; x++){
      // find the lambda and phi values appropriate for the pixel
      if(main->projection == CYLINDER){
        if(get_lam_phi(main, x, y, coords) == ERROR){
          fprintf(stderr,"lam, phi out of range for the (%d,%d) pixel\n", x, y);
        }
      }else if(main->projection == SATELLITE){
        if(get_lam_phi_satellite(main, x, y, coords, NULL) == ERROR){
          fprintf(stderr,"satellite projection: lam, phi out of range for the (%d,%d) pixel\n", x, y);
        }
      }else{
        fprintf(stderr,"invalid projection\n");
        return(ERROR);
      }
      // wrap the coordinates to be within the correct ranges for the data
      wrap_lam_phi(coords);
      // find the altitude for said lam, phi
      if(main->planet == EARTH){
        get_alt(coords, main->datapath, &alt);
        alt_to_rgb((double)alt, pixel);
      }else if(main->planet == MARS){
        get_alt_mars(coords, main->datapath, &alt);
        alt_to_rgb_mars((double)alt, pixel);
      }else{
        return(ERROR);
      }
      // set the pixel within the image to said colour
      if(bmp_set_pixel_rgb(bitmap, bmp, pixel, x, y)){
        fprintf(stderr,"failed to set RGB of pixel (%d,%d)\n",x,y);
        return(ERROR);
      }
    }
  }

  if(png_write(bitmap, bmp)){
    fprintf(stderr,"PNG output failed\n");
    return(ERROR);
  }

  return(OK);

}

int main(int argc, char **argv){

  int c=0;
  static int quiet_flag = 0;
  static int fit_flag = 0;
  static int earth_flag = 0;
  static int mars_flag = 0;
  static int satellite_flag = 0;
  static int help_flag = 0;
  main_t *main;
  struct stat filestat;

  if((main = (main_t *)malloc(sizeof(main_t))) == NULL){
    exit(EX_SOFTWARE);
  }

  // assign default variables
  main->width =         1000;
  main->height =        500;
  main->lam0 =          0.0;
  main->phi0 =          0.0;
  main->scale =         10000000.0;
  main->dpi =           72.0;
  main->datapath =      NULL;

  while(c>=0){
    static struct option long_options[] = {
      /* These options set a flag. */
      {"quiet",            no_argument,     &quiet_flag,   1},
      {"fit",              no_argument,       &fit_flag,   1},
      {"earth",            no_argument,     &earth_flag,   1},
      {"mars",             no_argument,      &mars_flag,   1},
      {"satellite",        no_argument, &satellite_flag,   1},
      {"help",             no_argument,      &help_flag,   1},
      /* These options don't set a flag.
         We distinguish them by their indices. */
      {"width",      required_argument,             0, 'w'},
      {"height",     required_argument,             0, 'h'},
      {"longitude",  required_argument,             0, 'l'},
      {"latitude",   required_argument,             0, 'p'},
      {"scale",      required_argument,             0, 's'},
      {"dpi",        required_argument,             0, 'd'},
      {"datapath",   required_argument,             0, 'i'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long(argc, argv, "qfemw:h:l:p:s:d:i:o:",
                    long_options, &option_index);

    switch(c){
      case -1:
        //end of options
        break;
      case 0:
        /* If this option set a flag, do nothing else now. */
        if(long_options[option_index].flag != 0){
          break;
        }
        break;
      case 'w':
        main->width = atol(optarg);
        break;
      case 'h':
        main->height = atol(optarg);
        break;
      case 'l':
        main->lam0 = atof(optarg);
        break;
      case 'p':
        main->phi0 = atof(optarg);
        break;
      case 's':
        main->scale = atof(optarg);
        break;
      case 'd':
        main->dpi = atof(optarg);
        break;
      case 'i':
        main->datapath = optarg;
        break;
      case '?':
        /* getopt_long already printed an error message. */
        break;
      default:
        exit(EX_USAGE);
    }
  }

  // Display usage information
  if(help_flag || (argc == 0)){
    printf("Usage: d2d <options> <PNG path>\n\n");
    printf("Options:\n");
    printf("  --width=<int>        The width of the image in pixels\n");
    printf("  --height=<int>       The height of the image in pixels\n");
    printf("  --longitude=<double> The longitude of the centre of the image\n");
    printf("  --latitude=<double>  The latitude of the centre of the image\n");
    printf("  --scale=<double>     The scale 1:n of the image to the planet surface\n");
    printf("  --dpi=<double>       The dpi of the screen on which the image is being viewed\n");
    printf("  --datapath=<string>  The path to the planet data (mandatory)\n");
    printf("  --fit                Adjusts the scale to produce an image of the entire planet\n");
    printf("                       (Note: overrides any set scale)\n");
    printf("  --earth              Choose predefined planetary values (default: Earth)\n  --mars\n");
    printf("  --satellite          Use the satellite projection\n");
    printf("  --quiet              Disable standard text output\n");
    printf("  --help               Display this help\n");
    printf("Example: ./d2d --satellite --earth --datapath=ETOPO2.raw.bin image.png\n");
    exit(EX_USAGE);
  }

  // Choose planetary parameter set
  if(earth_flag+mars_flag > 1){
    fprintf(stderr,"Choose only one set of planet parameters!\n");
    exit(EX_USAGE);
  }
  if(earth_flag+mars_flag == 0 ) {
    earth_flag=1;
  }

  if(earth_flag){
    main->rad_of_planet = 6378100.0;
    main->planet = EARTH;
  }
  if(mars_flag){
    main->rad_of_planet = 3397000.0;
    main->planet = MARS;
  }

  if(satellite_flag){
    main->projection = SATELLITE;
  }else{
    main->projection = CYLINDER;
  }

  // PNG filename handling
  if(optind > argc-1){
    fprintf(stderr,"Output specification missing\n");
    fprintf(stderr,"See the --help option for more details\n");
    exit(EX_USAGE);
  }
  if(optind < argc-1){
    fprintf(stderr,"Too many arguments\n");
    exit(EX_USAGE);
  }

  size_t len=strlen(argv[optind]);
  if((main->pngpath = (char *)malloc(len+1)) == NULL){
    exit(EX_USAGE);
  }
  strncpy(main->pngpath, argv[optind], len);
  main->pngpath[len+1]='\0';

  // check for user errors
  if(main->dpi < 0){
    exit(EX_USAGE);
  }
  if(main->width < 0){
    exit(EX_USAGE);
  }
  if(main->height < 0){
    exit(EX_USAGE);
  }

  if(fit_flag){
    switch(main->projection){
    case CYLINDER:
      if(main->width >= 2*main->height){
        main->scale = M_PI*main->rad_of_planet*(double)main->dpi/(double)(main->height*INS_TO_M);
      }else{
        main->scale = 2.0*M_PI*main->rad_of_planet*(double)main->dpi/(double)(main->width*INS_TO_M);
      }
      break;
    case SATELLITE:
      if(main->width > main->height){
        main->scale = 2.0*main->rad_of_planet*(double)main->dpi/(double)(main->height*INS_TO_M);
      }else{
        main->scale = 2.0*main->rad_of_planet*(double)main->dpi/(double)(main->width*INS_TO_M);
      }
      break;
    default:
      fprintf(stderr,"Invalid projection\n");
      exit(EX_USAGE);
    }
  }
  if(main->scale < 0){
    exit(EX_USAGE);
  }

  main->deg_per_pix= 180.0*INS_TO_M*main->scale/(main->dpi*M_PI*main->rad_of_planet);

  if(main->planet == EARTH){
    if(stat(main->datapath, &filestat) == -1){
      print_errno_info(errno);
      fprintf(stderr,"Path to source data missing or invalid\n");
      fprintf(stderr,"See the --help option for more details\n");
      exit(EX_USAGE);
    }
  }

  if(!quiet_flag){
    // Print all options set
    printf("Druantia 2D - A program to generate a 2D colour map\n");
    printf("              from altitude data\n");
    printf("Copyright (C) 2005 - 2006\n");
    printf("              Joern Wilms, Robert Dowse, Robert Swain\n\n");
    printf("Image width         = %d\n", main->width);
    printf("Image height        = %d\n", main->height);
    printf("Centre longitude    = %lf\n", main->lam0);
    printf("Centre latitude     = %lf\n", main->phi0);
    printf("Scale               = %lf\n", main->scale);
    printf("DPI                 = %lf\n", main->dpi);
    printf("Path to data        = %s\n", main->datapath);
    if(!strcmp(main->pngpath,"-")){
      printf("Image output        = stdout\n");
    }else{
      printf("Image output        = %s\n", main->pngpath);
    }
    if(earth_flag){
      printf("Earth parameters chosen\n");
    }
    if(mars_flag){
      printf("Mars parameters chosen\n");
    }
  }

  if(druantia2d(main) == ERROR){
    fprintf(stderr,"error in druantia2d.\n");
    exit(EX_SOFTWARE);
  }

  return 0;

}
