/* druantia3d.cpp - A program to generate a 3D visualisation of a planet
 *                  using OpenSceneGraph
 *
 * Copyright (C) 2005 - 2006
 *                    Joern Wilms
 *                    Robert Dowse
 *                    Robert Swain
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

extern "C" {
#include "../druantia.h"
#include "../common/common.h"
#include <errno.h>
#include <getopt.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
};

#include <osg/Geode>
#include <osg/Group>
#include <osg/ShapeDrawable>
#include <osg/Image>
#include <osg/Texture2D>
#include <osgUtil/Optimizer>
#include <osg/PositionAttitudeTransform>
#include <osgProducer/Viewer>
#include <osgGA/UFOManipulator>

#define WINDOW_WIDTH  512
#define WINDOW_HEIGHT 512
#define FIELD_SIZE_POWER 9

int druantia3d(main_t *main){

  // Declare some variables, pointers and structures
  short alt;
  int x, y, counter;
  unsigned int manipulatorNumber;
  double cam_h;
  uint8_t *bmp;
  double *cart;
  bitmap_t *bmpinfo;
  pixel_t *colour;
  coords_t *coords;
  horizon_t *horizon, *horizon_root;

  // Allocate the structures
  if((coords = (coords_t *)malloc(sizeof(coords_t))) == NULL){
    return(ERROR);
  }
  if((colour = (pixel_t *)malloc(sizeof(pixel_t))) == NULL){
    return(ERROR);
  }
  if((bmpinfo = (bitmap_t *)malloc(sizeof(bitmap_t))) == NULL){
    return(ERROR);
  }
  if((cart = (double *)malloc(sizeof(double)*3)) == NULL){
    return(ERROR);
  }
  if((horizon = (horizon_t *)malloc(sizeof(horizon_t))) == NULL){
    return(ERROR);
  }
  if((horizon_root = (horizon_t *)malloc(sizeof(horizon_t))) == NULL){
    return(ERROR);
  }

  // Set the default values
  // Note that some of these are not defaults and will need moving to function parameters
  bmpinfo->width =    main->width;
  bmpinfo->height =   main->height;
  main->deg_per_pix=    180.0*INS_TO_M*main->scale/(main->dpi*M_PI*main->rad_of_planet);

  // Create a root group to contain the scene
  osg::Group* root = new osg::Group();
  // Create a new Geode
  // Only classes derived from the Node class can be rendered
  osg::Geode* heightFieldGeode = new osg::Geode();
  // Create a new heightfield
  osg::HeightField* heightField = new osg::HeightField;
  // Create a new drawable using the heightField
  osg::ShapeDrawable* heightFieldDrawable = new osg::ShapeDrawable(heightField);
  // Create an image
  osg::Image* heightFieldImage = new osg::Image;
  // Create a 2D texture
  osg::Texture2D* heightFieldTexture = new osg::Texture2D;
  // Create a new StateSet
  osg::StateSet* heightFieldState = new osg::StateSet();
  // Create non-full screen render surface (a window)
  osg::ref_ptr<Producer::RenderSurface> window (new Producer::RenderSurface);
  // Create a new camera
  osg::ref_ptr<Producer::Camera> camera1 (new Producer::Camera);
  // Create a new camera configuration
  osg::ref_ptr<Producer::CameraConfig> camera1Config (new Producer::CameraConfig());
  // Create a UFOManipulator as this appears to be the most appropriate camera control method
  osgGA::UFOManipulator* UFOManipulator = new osgGA::UFOManipulator;

  // Initialise and allocate the heightField
  heightField->allocate(main->width,main->height);
  // Set the X and Y interval size
  // I don't actually know what this value does so we'll leave it as 1.0f
  heightField->setXInterval(1.0f);
  heightField->setYInterval(1.0f);
  // Initialise the bitmap texture data
  if((bmp = bmp_init(bmpinfo)) == NULL){
    printf("Bitmap initialisation failed\n");
    return(ERROR);
  }

  // Create a loop over the rows, then columns of the array
  for(y=0; y<main->height; y++){
    for(x=0; x<main->width; x++){
      // Find the lambda and phi values corresponding to the (i,j)th position
      if(get_lam_phi_satellite(main, x, y, coords, cart)){
        printf("lam, phi out of range for the (%d,%d) pixel",x,y);
      }
      // wrap the coordinates to be within the correct ranges for the data
      wrap_lam_phi(coords);
      if(main->planet == EARTH){
        // Get the altitude from the data for said (lam,phi)
        get_alt(coords, main->datapath, &alt);
        // Convert the altitude into a colour
        alt_to_rgb((double)alt, colour);
      }else if(main->planet == MARS){
        // Get the altitude from the data for said (lam,phi)
        get_alt_mars(coords, main->datapath, &alt);
        // Convert the altitude into a colour
        alt_to_rgb_mars((double)alt, colour);
      }else{
        return(ERROR);
      }
      // Set the appropriate pixel colour in the bitmap
      if(bmp_set_pixel_rgb(bmpinfo,bmp,colour,x,y)){
        printf("Failed to set RGB of pixel (%d,%d)\n",x,y);
        return(ERROR);
      }
      if(alt == -32767){cart[0] = (int)32767*main->alt_factor;}
      // Set the height in the heightfield array
      heightField->setHeight(x,y,(float)((main->alt_factor*alt + cart[0])*main->dpi/(main->scale*INS_TO_M)));
    }
  }

  // Print a helpful note to show when the array has been populated
  printf("Done filling the array.\n");

  // Set the source of the image data to our bitmap and set its properties
  heightFieldImage->setImage(bmpinfo->width,bmpinfo->height,1,GL_RGB,GL_RGB,
    GL_UNSIGNED_BYTE,bmp,osg::Image::USE_MALLOC_FREE,1);
  // Set the texture dimensions
  heightFieldTexture->setTextureSize(bmpinfo->width,bmpinfo->height);
  // Set the texture image source
  heightFieldTexture->setImage(heightFieldImage);
  // Protect the texture from being omitted due to optimisation as a static state
  heightFieldTexture->setDataVariance(osg::Object::DYNAMIC);
  // Assign texture unit 0 of our new StateSet to the texture and enable it
  heightFieldState->setTextureAttributeAndModes(0,heightFieldTexture,
    osg::StateAttribute::ON);

  // Associate this StateSet with the geode that contains
  // the object
  heightFieldGeode->setStateSet(heightFieldState);
  // Add the heightFieldDrawable as a drawable to the geode
  heightFieldGeode->addDrawable(heightFieldDrawable);
  // Add the heightFieldGeode as a child of the root scene
  root->addChild(heightFieldGeode);

  // Set the screen in which the window should display
  window->setScreenNum(0);
  // Set the size of the window
  window->setWindowRectangle(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
  // Set the window title
  window->setWindowName ("Druantia 3D");
  // Link the window to the camera
  camera1->setRenderSurface(window.get());
  // Set the camera view size
  camera1->setProjectionRectangle(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
  // Link the configuration to the camera
  camera1Config->addCamera("The only camera", camera1.get());

  // Create a Producer-based viewer using the camera defined above
  osgProducer::Viewer viewer(camera1Config.get());

  // Set up the viewer with the standard settings
  viewer.setUpViewer(osgProducer::Viewer::STANDARD_SETTINGS);

  // Link the UFOManipulator to the viewer and select it to manipulate the camera
  manipulatorNumber = viewer.addCameraManipulator(UFOManipulator);
  viewer.selectCameraManipulator(manipulatorNumber);
  // Set the initial/home position of the camera
  UFOManipulator->setHomePosition(
      osg::Vec3(main->width/2,main->height/2-1,
                (main->rad_of_planet + main->alt_factor*main->cam_height)*main->dpi/(main->scale*INS_TO_M)), // eye
//  osg::Vec3(main->width/2,main->height/2,0), // centre
      osg::Vec3(main->width/2,main->height/2,
                (main->rad_of_planet + main->alt_factor*main->cam_height)*main->dpi/(main->scale*INS_TO_M)), // centre
      osg::Vec3(0.0,0.0,1.0),false); // up

  // Optimise the scene
  osgUtil::Optimizer optimizer;
  optimizer.optimize(root);

  // Set the scene data to be rendered by the viewer
  viewer.setSceneData(root);
  // Create the window and render surface and render the scene
  viewer.realize();

  viewer.sync();
  viewer.update();
  viewer.frame();
  cam_h = xyz_to_height(main, viewer.getPosition());
  get_horizon(main, cam_h, horizon_root);
  std::cout << "horizon_root = " << horizon_root->l1 << "\n";

  const double *_curpos;
  double xdisp, ydisp;

  // Update the scene within the viewer until done
  while(!viewer.done()){
    viewer.sync();
    viewer.update();

    if(!(counter%1000)){
      cam_h = xyz_to_height(main, _curpos = viewer.getPosition());
      xdisp = _curpos[0] - (main->width>>1);
      ydisp = _curpos[1] - (main->height>>1);
      get_horizon(main, cam_h, horizon);
      std::cout << "horizon = " << horizon->l1 << "\n";
      if( horizon->l1 > 1.25*horizon_root->l1 || // for zoom out
          horizon->l1 < 0.75*horizon_root->l1 || // for zoom in
                                                  // for x,y motion
          sqrt(xdisp*xdisp + ydisp*ydisp)*INS_TO_M*main->scale/(double)main->dpi > 0.1*horizon_root->l2){
        std::cout << "TRIGGER!\n";
      } else
        std::cout << "Within range.\n";
    }
    counter++;

    viewer.frame();
  }

  viewer.sync();
  return(OK);

}



int main(int argc, char **argv){

  int c=0;
  static int quiet_flag = 0;
  static int fit_flag = 0;
  static int datalimit_flag = 0;
  static int earth_flag = 0;
  static int mars_flag = 0;
  static int help_flag = 0;
  main_t *main;
  horizon_t *horizon_init;
  struct stat filestat;

  if((main = (main_t *)malloc(sizeof(main_t))) == NULL){
    exit(EX_SOFTWARE);
  }
  if((horizon_init = (horizon_t *)malloc(sizeof(horizon_t))) == NULL){
    return(ERROR);
  }

  // assign default variables
  main->width =         (int)pow(2,FIELD_SIZE_POWER);
  main->height =        (int)pow(2,FIELD_SIZE_POWER);
  main->cam_height =    50000.0;
  main->alt_factor =    1.0;
  main->lam0 =          0.0;
  main->phi0 =          0.0;
  main->dpi =           72.0;
  main->datapath =      NULL;

  while(c>=0){
    static struct option long_options[] = {
      /* These options set a flag. */
      {"quiet",            no_argument,     &quiet_flag,   1},
      {"fit",              no_argument,       &fit_flag,   1},
      {"datalimit",        no_argument, &datalimit_flag,   1},
      {"earth",            no_argument,     &earth_flag,   1},
      {"mars",             no_argument,      &mars_flag,   1},
      {"help",             no_argument,      &help_flag,   1},
      /* These options don't set a flag.
         We distinguish them by their indices. */
      {"width",      required_argument,             0, 'w'},
      {"height",     required_argument,             0, 'h'},
      {"longitude",  required_argument,             0, 'l'},
      {"latitude",   required_argument,             0, 'p'},
      {"camheight",  required_argument,             0, 'c'},
      {"altfactor",  required_argument,             0, 'a'},
      {"dpi",        required_argument,             0, 'd'},
      {"datapath",   required_argument,             0, 'i'},
      {0, 0, 0, 0}
    };
    /* getopt_long stores the option index here. */
    int option_index = 0;

    c = getopt_long(argc, argv, "qftemw:h:l:p:c:a:d:i:", long_options, &option_index);

    switch(c){
      case -1:
        //end of options
        break;
      case 0:
        /* If this option set a flag, do nothing else now. */
        if(long_options[option_index].flag != 0){
          break;
        }
        break;
      case 'w':
        main->width = atol(optarg);
        break;
      case 'h':
        main->height = atol(optarg);
        break;
      case 'l':
        main->lam0 = atof(optarg);
        break;
      case 'p':
        main->phi0 = atof(optarg);
        break;
      case 'c':
        main->cam_height = atof(optarg);
        break;
      case 'a':
        main->alt_factor = atof(optarg);
        break;
      case 'd':
        main->dpi = atof(optarg);
        break;
      case 'i':
        main->datapath = optarg;
        break;
      case '?':
        /* getopt_long already printed an error message. */
        break;
      default:
        exit(EX_USAGE);
    }
  }

  // Display usage information
  if(help_flag || (argc == 0)){
    printf("Usage: d3d <options>\n\n");
    printf("Options:\n");
    printf("  --width=<int>\tThe width of the image in pixels\n");
    printf("  --height=<int>\tThe height of the image in pixels\n");
    printf("  --longitude=<double>\tThe longitude of the centre of the image\n");
    printf("  --latitude=<double>\tThe latitude of the centre of the image\n");
    printf("  --camheight=<double>\tThe height of the camera above the ground\n");
    printf("  --altfactor=<double>\tExaggeration factor of altitude offsets from 'sea level'\n");
    printf("  --dpi=<double>\tThe dpi of the screen on which the image is being viewed\n");
    printf("  --datapath=<string>\tThe path to the planet data (mandatory for Earth)\n");
    printf("  --datalimit\t\t\tCalculate the scale based on the data resolution at the equator\n");
    printf("  \t\t\t(Note: overrides the camera height)\n");
    printf("  --fit\t\t\tAdjusts the scale to produce an image of the entire planet\n");
    printf("  \t\t\t(Note: overrides the camera height and data resolution limit)\n");
    printf("  --earth\t\tChoose predefined planetary values (default: Earth)\n  --mars\n");
    printf("  --quiet\t\tDisable standard text output\n");
    printf("  --help\t\tDisplay this help\n");
    printf("Example: ./d3d --earth --datapath=ETOPO2.raw.bin\n");
    exit(EX_USAGE);
  }

  // Choose planetary parameter set
  if(earth_flag+mars_flag > 1){
    fprintf(stderr,"Choose only one set of planet parameters!\n");
    exit(EX_USAGE);
  }
  if(earth_flag+mars_flag == 0 ) {
    earth_flag=1;
  }

  if(earth_flag){
    main->rad_of_planet = 6378100.0;
    main->planet = EARTH;
  }
  if(mars_flag){
    main->rad_of_planet = 3396000.0;
    main->planet = MARS;
  }

  // check for user errors
  if(main->dpi < 0){
    exit(EX_USAGE);
  }
  if(main->width < 0){
    exit(EX_USAGE);
  }
  if(main->height < 0){
    exit(EX_USAGE);
  }

  if(fit_flag){
    if(main->width > main->height){
      main->scale = 2.0*main->rad_of_planet*main->dpi/(main->height*INS_TO_M);
    }else{
      main->scale = 2.0*main->rad_of_planet*main->dpi/(main->width*INS_TO_M);
    }
  }else if(datalimit_flag){
    if(mars_flag){
      main->scale = 463.057667*(double)main->dpi/INS_TO_M; // MOLA data resolution at equator
    }else if(earth_flag){
      main->scale = 3710.62817*(double)main->dpi/INS_TO_M; // ETOPO-2 data resolution at equator
    }
  }else{
    // calculate the scale according to the camera height
    get_horizon(main, main->cam_height, horizon_init);
    main->scale = 2.0*horizon_init->l2*main->dpi/(main->width*INS_TO_M); // half the field
    free(horizon_init);
  }

  if(main->scale < 0){
    exit(EX_USAGE);
  }

  if(main->planet == EARTH){
    if(stat(main->datapath, &filestat) == -1){
      print_errno_info(errno);
      fprintf(stderr,"path to source data missing or invalid\n");
      fprintf(stderr,"see the --help option for more details\n");
      exit(EX_USAGE);
    }
  }

  if(!quiet_flag){
    // Print all options set
    printf("Druantia 3D - A program to generate a 3D visualisation of a planet\n");
    printf("              from altitude data\n");
    printf("Copyright (C) 2005 - 2006\n");
    printf("              Joern Wilms, Robert Dowse, Robert Swain\n\n");
    printf("Image width         = %d\n",  main->width);
    printf("Image height        = %d\n",  main->height);
    printf("Centre longitude    = %lf\n", main->lam0);
    printf("Centre latitude     = %lf\n", main->phi0);
    printf("Camera height       = %lf\n", main->cam_height);
    printf("DPI                 = %lf\n", main->dpi);
    if(earth_flag){
      printf("Earth parameters chosen\n\n");
    }
    if(mars_flag){
      printf("Mars parameters chosen\n\n");
    }
  }

  if(druantia3d(main) == ERROR){
    fprintf(stderr,"*** ERROR in Druantia 3D ***\n");
    exit(EX_SOFTWARE);
  }

  return 0;

}
